@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 2010, 2011, 2012, 2013, 2014, 2016 Nicira, Inc.
+ * Copyright (c) 2010, 2011, 2012, 2013, 2014 Nicira, Inc.
+ * Copyright (c) 2016 Red Hat, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -27,86 +28,44 @@
 #include <sys/ioctl.h>
 
 #include "byte-order.h"
-#include "csum.h"
 #include "daemon.h"
 #include "dirs.h"
 #include "dpif.h"
-#include "dp-packet.h"
-#include "dynamic-string.h"
-#include "flow.h"
-#include "hash.h"
-#include "hmap.h"
-#include "list.h"
+#include "netdev.h"
+#include "netdev-native-tnl.h"
 #include "netdev-provider.h"
-#include "odp-netlink.h"
-#include "dp-packet.h"
+#include "netdev-vport-private.h"
 #include "ovs-router.h"
 #include "packets.h"
 #include "poll-loop.h"
 #include "route-table.h"
-#include "shash.h"
+#include "smap.h"
 #include "socket-util.h"
-#include "openvswitch/vlog.h"
 #include "unaligned.h"
 #include "unixctl.h"
-#include "util.h"
+#include "openvswitch/vlog.h"
 
 VLOG_DEFINE_THIS_MODULE(netdev_vport);
-static struct vlog_rate_limit err_rl = VLOG_RATE_LIMIT_INIT(60, 5);
 
 #define GENEVE_DST_PORT 6081
 #define VXLAN_DST_PORT 4789
 #define LISP_DST_PORT 4341
 #define STT_DST_PORT 7471
 
-#define VXLAN_HLEN   (sizeof(struct udp_header) +         \
-                      sizeof(struct vxlanhdr))
-
-#define GENEVE_BASE_HLEN   (sizeof(struct udp_header) +         \
-                            sizeof(struct genevehdr))
-
 #define DEFAULT_TTL 64
 
-struct netdev_vport {
-    struct netdev up;
+/* Last read of the route-table's change number. */
+static uint64_t rt_change_seqno;
 
-    /* Protects all members below. */
-    struct ovs_mutex mutex;
-
-    struct eth_addr etheraddr;
-    struct netdev_stats stats;
-
-    /* Tunnels. */
-    struct netdev_tunnel_config tnl_cfg;
-    char egress_iface[IFNAMSIZ];
-    bool carrier_status;
-
-    /* Patch Ports. */
-    char *peer;
-};
+static int get_patch_config(const struct netdev *netdev, struct smap *args);
+static int get_tunnel_config(const struct netdev *, struct smap *args);
+static bool tunnel_check_status_change__(struct netdev_vport *);
 
 struct vport_class {
     const char *dpif_port;
     struct netdev_class netdev_class;
 };
 
-/* Last read of the route-table's change number. */
-static uint64_t rt_change_seqno;
-
-static int netdev_vport_construct(struct netdev *);
-static int get_patch_config(const struct netdev *netdev, struct smap *args);
-static int get_tunnel_config(const struct netdev *, struct smap *args);
-static bool tunnel_check_status_change__(struct netdev_vport *);
-
-static uint16_t tnl_udp_port_min = 32768;
-static uint16_t tnl_udp_port_max = 61000;
-
-static bool
-is_vport_class(const struct netdev_class *class)
-{
-    return class->construct == netdev_vport_construct;
-}
-
 bool
 netdev_vport_is_vport_class(const struct netdev_class *class)
 {
@@ -120,13 +79,6 @@ vport_class_cast(const struct netdev_class *class)
     return CONTAINER_OF(class, struct vport_class, netdev_class);
 }
 
-static struct netdev_vport *
-netdev_vport_cast(const struct netdev *netdev)
-{
-    ovs_assert(is_vport_class(netdev_get_class(netdev)));
-    return CONTAINER_OF(netdev, struct netdev_vport, up);
-}
-
 static const struct netdev_tunnel_config *
 get_netdev_tunnel_config(const struct netdev *netdev)
 {
@@ -195,15 +147,6 @@ netdev_vport_get_dpif_port(const struct netdev *netdev,
     }
 }
 
-char *
-netdev_vport_get_dpif_port_strdup(const struct netdev *netdev)
-{
-    char namebuf[NETDEV_VPORT_NAME_BUFSIZE];
-
-    return xstrdup(netdev_vport_get_dpif_port(netdev, namebuf,
-                                              sizeof namebuf));
-}
-
 /* Whenever the route-table change number is incremented,
  * netdev_vport_route_changed() should be called to update
  * the corresponding tunnel interface status. */
@@ -240,7 +183,7 @@ netdev_vport_alloc(void)
     return &netdev->up;
 }
 
-static int
+int
 netdev_vport_construct(struct netdev *netdev_)
 {
     struct netdev_vport *dev = netdev_vport_cast(netdev_);
@@ -319,10 +262,10 @@ tunnel_check_status_change__(struct netdev_vport *netdev)
 
     iface[0] = '\0';
     route = &netdev->tnl_cfg.ipv6_dst;
-    if (ovs_router_lookup(route, iface, &gw)) {
+    if (ovs_router_lookup(route, iface, NULL, &gw)) {
         struct netdev *egress_netdev;
 
-        if (!netdev_open(iface, "system", &egress_netdev)) {
+        if (!netdev_open(iface, NULL, &egress_netdev)) {
             status = netdev_get_carrier(egress_netdev);
             netdev_close(egress_netdev);
         }
@@ -369,7 +312,7 @@ netdev_vport_update_flags(struct netdev *netdev OVS_UNUSED,
 }
 
 static void
-netdev_vport_run(void)
+netdev_vport_run(const struct netdev_class *netdev_class OVS_UNUSED)
 {
     uint64_t seq;
 
@@ -382,7 +325,7 @@ netdev_vport_run(void)
 }
 
 static void
-netdev_vport_wait(void)
+netdev_vport_wait(const struct netdev_class *netdev_class OVS_UNUSED)
 {
     uint64_t seq;
 
@@ -504,10 +447,6 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args)
                           name, node->value);
                 return EINVAL;
             }
-            if (dst_proto == ETH_TYPE_IPV6) {
-                VLOG_WARN("%s: IPv6 'remote_ip' is not supported", name);
-                return EOPNOTSUPP;
-            }
         } else if (!strcmp(node->key, "local_ip")) {
             int err;
             err = parse_tunnel_ip(node->value, true, &tnl_cfg.ip_src_flow,
@@ -517,10 +456,6 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args)
                 VLOG_WARN("%s: bad %s 'local_ip'", name, type);
                 break;
             }
-            if (src_proto == ETH_TYPE_IPV6) {
-                VLOG_WARN("%s: IPv6 'local_ip' is not supported", name);
-                return EOPNOTSUPP;
-            }
         } else if (!strcmp(node->key, "tos")) {
             if (!strcmp(node->value, "inherit")) {
                 tnl_cfg.tos_inherit = true;
@@ -608,6 +543,8 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args)
         static struct ovs_mutex mutex = OVS_MUTEX_INITIALIZER;
         static pid_t pid = 0;
 
+        VLOG_ERR("%s: OVS IPsec tunnel support is deprecated.", name);
+
 #ifndef _WIN32
         ovs_mutex_lock(&mutex);
         if (pid <= 0) {
@@ -859,622 +796,16 @@ get_stats(const struct netdev *netdev, struct netdev_stats *stats)
     struct netdev_vport *dev = netdev_vport_cast(netdev);
 
     ovs_mutex_lock(&dev->mutex);
-    *stats = dev->stats;
+    /* Passing only collected counters */
+    stats->tx_packets = dev->stats.tx_packets;
+    stats->tx_bytes = dev->stats.tx_bytes;
+    stats->rx_packets = dev->stats.rx_packets;
+    stats->rx_bytes = dev->stats.rx_bytes;
     ovs_mutex_unlock(&dev->mutex);
 
     return 0;
 }
 
-
-/* Tunnel push pop ops. */
-
-static struct ip_header *
-ip_hdr(void *eth)
-{
-    return (void *)((char *)eth + sizeof (struct eth_header));
-}
-
-static struct ovs_16aligned_ip6_hdr *
-ipv6_hdr(void *eth)
-{
-    return (void *)((char *)eth + sizeof (struct eth_header));
-}
-
-static void *
-ip_extract_tnl_md(struct dp_packet *packet, struct flow_tnl *tnl,
-                  unsigned int *hlen)
-{
-    void *nh;
-    struct ip_header *ip;
-    struct ovs_16aligned_ip6_hdr *ip6;
-    void *l4;
-    int l3_size;
-
-    nh = dp_packet_l3(packet);
-    ip = nh;
-    ip6 = nh;
-    l4 = dp_packet_l4(packet);
-
-    if (!nh || !l4) {
-        return NULL;
-    }
-
-    *hlen = sizeof(struct eth_header);
-
-    l3_size = dp_packet_size(packet) -
-              ((char *)nh - (char *)dp_packet_data(packet));
-
-    if (IP_VER(ip->ip_ihl_ver) == 4) {
-
-        ovs_be32 ip_src, ip_dst;
-
-        if (csum(ip, IP_IHL(ip->ip_ihl_ver) * 4)) {
-            VLOG_WARN_RL(&err_rl, "ip packet has invalid checksum");
-            return NULL;
-        }
-
-        if (ntohs(ip->ip_tot_len) > l3_size) {
-            VLOG_WARN_RL(&err_rl, "ip packet is truncated (IP length %d, actual %d)",
-                         ntohs(ip->ip_tot_len), l3_size);
-            return NULL;
-        }
-        if (IP_IHL(ip->ip_ihl_ver) * 4 > sizeof(struct ip_header)) {
-            VLOG_WARN_RL(&err_rl, "ip options not supported on tunnel packets "
-                         "(%d bytes)", IP_IHL(ip->ip_ihl_ver) * 4);
-            return NULL;
-        }
-
-        ip_src = get_16aligned_be32(&ip->ip_src);
-        ip_dst = get_16aligned_be32(&ip->ip_dst);
-
-        tnl->ip_src = ip_src;
-        tnl->ip_dst = ip_dst;
-        tnl->ip_tos = ip->ip_tos;
-        tnl->ip_ttl = ip->ip_ttl;
-
-        *hlen += IP_HEADER_LEN;
-
-    } else if (IP_VER(ip->ip_ihl_ver) == 6) {
-
-        memcpy(tnl->ipv6_src.s6_addr, ip6->ip6_src.be16, sizeof ip6->ip6_src);
-        memcpy(tnl->ipv6_dst.s6_addr, ip6->ip6_dst.be16, sizeof ip6->ip6_dst);
-        tnl->ip_tos = 0;
-        tnl->ip_ttl = ip6->ip6_hlim;
-
-        *hlen += IPV6_HEADER_LEN;
-
-    } else {
-        VLOG_WARN_RL(&err_rl, "ipv4 packet has invalid version (%d)",
-                     IP_VER(ip->ip_ihl_ver));
-        return NULL;
-    }
-
-    return l4;
-}
-
-static bool
-is_header_ipv6(const void *header)
-{
-    const struct eth_header *eth;
-    eth = header;
-    return eth->eth_type == htons(ETH_TYPE_IPV6);
-}
-
-/* Pushes the 'size' bytes of 'header' into the headroom of 'packet',
- * reallocating the packet if necessary.  'header' should contain an Ethernet
- * header, followed by an IPv4 header (without options), and an L4 header.
- *
- * This function sets the IP header's ip_tot_len field (which should be zeroed
- * as part of 'header') and puts its value into '*ip_tot_size' as well.  Also
- * updates IP header checksum.
- *
- * Return pointer to the L4 header added to 'packet'. */
-static void *
-push_ip_header(struct dp_packet *packet,
-               const void *header, int size, int *ip_tot_size)
-{
-    struct eth_header *eth;
-    struct ip_header *ip;
-    struct ovs_16aligned_ip6_hdr *ip6;
-
-    eth = dp_packet_push_uninit(packet, size);
-    *ip_tot_size = dp_packet_size(packet) - sizeof (struct eth_header);
-
-    memcpy(eth, header, size);
-
-    if (is_header_ipv6(header)) {
-        ip6 = ipv6_hdr(eth);
-        *ip_tot_size -= IPV6_HEADER_LEN;
-        ip6->ip6_plen = htons(*ip_tot_size);
-        return ip6 + 1;
-    } else {
-        ip = ip_hdr(eth);
-        ip->ip_tot_len = htons(*ip_tot_size);
-        ip->ip_csum = recalc_csum16(ip->ip_csum, 0, ip->ip_tot_len);
-        *ip_tot_size -= IP_HEADER_LEN;
-        return ip + 1;
-    }
-}
-
-static void *
-udp_extract_tnl_md(struct dp_packet *packet, struct flow_tnl *tnl,
-                   unsigned int *hlen)
-{
-    struct udp_header *udp;
-
-    udp = ip_extract_tnl_md(packet, tnl, hlen);
-    if (!udp) {
-        return NULL;
-    }
-
-    if (udp->udp_csum) {
-        uint32_t csum;
-        if (is_header_ipv6(dp_packet_data(packet))) {
-            csum = packet_csum_pseudoheader6(dp_packet_l3(packet));
-        } else {
-            csum = packet_csum_pseudoheader(dp_packet_l3(packet));
-        }
-
-        csum = csum_continue(csum, udp, dp_packet_size(packet) -
-                             ((const unsigned char *)udp -
-                              (const unsigned char *)dp_packet_l2(packet)));
-        if (csum_finish(csum)) {
-            return NULL;
-        }
-        tnl->flags |= FLOW_TNL_F_CSUM;
-    }
-
-    tnl->tp_src = udp->udp_src;
-    tnl->tp_dst = udp->udp_dst;
-
-    return udp + 1;
-}
-
-static ovs_be16
-get_src_port(struct dp_packet *packet)
-{
-    uint32_t hash;
-
-    hash = dp_packet_get_rss_hash(packet);
-
-    return htons((((uint64_t) hash * (tnl_udp_port_max - tnl_udp_port_min)) >> 32) +
-                 tnl_udp_port_min);
-}
-
-static void
-push_udp_header(struct dp_packet *packet,
-                const struct ovs_action_push_tnl *data)
-{
-    struct udp_header *udp;
-    int ip_tot_size;
-
-    udp = push_ip_header(packet, data->header, data->header_len, &ip_tot_size);
-
-    /* set udp src port */
-    udp->udp_src = get_src_port(packet);
-    udp->udp_len = htons(ip_tot_size);
-
-    if (udp->udp_csum) {
-        uint32_t csum;
-        if (is_header_ipv6(dp_packet_data(packet))) {
-            csum = packet_csum_pseudoheader6(ipv6_hdr(dp_packet_data(packet)));
-        } else {
-            csum = packet_csum_pseudoheader(ip_hdr(dp_packet_data(packet)));
-        }
-
-        csum = csum_continue(csum, udp, ip_tot_size);
-        udp->udp_csum = csum_finish(csum);
-
-        if (!udp->udp_csum) {
-            udp->udp_csum = htons(0xffff);
-        }
-    }
-}
-
-static void *
-udp_build_header(struct netdev_tunnel_config *tnl_cfg,
-                 const struct flow *tnl_flow,
-                 struct ovs_action_push_tnl *data,
-                 unsigned int *hlen)
-{
-    struct ip_header *ip;
-    struct ovs_16aligned_ip6_hdr *ip6;
-    struct udp_header *udp;
-    bool is_ipv6;
-
-    *hlen = sizeof(struct eth_header);
-
-    is_ipv6 = is_header_ipv6(data->header);
-
-    if (is_ipv6) {
-        ip6 = ipv6_hdr(data->header);
-        ip6->ip6_nxt = IPPROTO_UDP;
-        udp = (struct udp_header *) (ip6 + 1);
-        *hlen += IPV6_HEADER_LEN;
-    } else {
-        ip = ip_hdr(data->header);
-        ip->ip_proto = IPPROTO_UDP;
-        udp = (struct udp_header *) (ip + 1);
-        *hlen += IP_HEADER_LEN;
-    }
-
-    udp->udp_dst = tnl_cfg->dst_port;
-
-    if (is_ipv6 || tnl_flow->tunnel.flags & FLOW_TNL_F_CSUM) {
-        /* Write a value in now to mark that we should compute the checksum
-         * later. 0xffff is handy because it is transparent to the
-         * calculation. */
-        udp->udp_csum = htons(0xffff);
-    }
-
-    return udp + 1;
-}
-
-static int
-gre_header_len(ovs_be16 flags)
-{
-    int hlen = 4;
-
-    if (flags & htons(GRE_CSUM)) {
-        hlen += 4;
-    }
-    if (flags & htons(GRE_KEY)) {
-        hlen += 4;
-    }
-    if (flags & htons(GRE_SEQ)) {
-        hlen += 4;
-    }
-    return hlen;
-}
-
-static int
-parse_gre_header(struct dp_packet *packet,
-                 struct flow_tnl *tnl)
-{
-    const struct gre_base_hdr *greh;
-    ovs_16aligned_be32 *options;
-    int hlen;
-    unsigned int ulen;
-
-    greh = ip_extract_tnl_md(packet, tnl, &ulen);
-    if (!greh) {
-        return -EINVAL;
-    }
-
-    if (greh->flags & ~(htons(GRE_CSUM | GRE_KEY | GRE_SEQ))) {
-        return -EINVAL;
-    }
-
-    if (greh->protocol != htons(ETH_TYPE_TEB)) {
-        return -EINVAL;
-    }
-
-    hlen = ulen + gre_header_len(greh->flags);
-    if (hlen > dp_packet_size(packet)) {
-        return -EINVAL;
-    }
-
-    options = (ovs_16aligned_be32 *)(greh + 1);
-    if (greh->flags & htons(GRE_CSUM)) {
-        ovs_be16 pkt_csum;
-
-        pkt_csum = csum(greh, dp_packet_size(packet) -
-                              ((const unsigned char *)greh -
-                               (const unsigned char *)dp_packet_l2(packet)));
-        if (pkt_csum) {
-            return -EINVAL;
-        }
-        tnl->flags = FLOW_TNL_F_CSUM;
-        options++;
-    }
-
-    if (greh->flags & htons(GRE_KEY)) {
-        tnl->tun_id = be32_to_be64(get_16aligned_be32(options));
-        tnl->flags |= FLOW_TNL_F_KEY;
-        options++;
-    }
-
-    if (greh->flags & htons(GRE_SEQ)) {
-        options++;
-    }
-
-    return hlen;
-}
-
-static void
-pkt_metadata_init_tnl(struct pkt_metadata *md)
-{
-    /* Zero up through the tunnel metadata options. The length and table
-     * are before this and as long as they are empty, the options won't
-     * be looked at. */
-    memset(md, 0, offsetof(struct pkt_metadata, tunnel.metadata.opts));
-}
-
-static int
-netdev_gre_pop_header(struct dp_packet *packet)
-{
-    struct pkt_metadata *md = &packet->md;
-    struct flow_tnl *tnl = &md->tunnel;
-    int hlen = sizeof(struct eth_header) + 4;
-
-    hlen += is_header_ipv6(dp_packet_data(packet)) ?
-            IPV6_HEADER_LEN : IP_HEADER_LEN;
-
-    pkt_metadata_init_tnl(md);
-    if (hlen > dp_packet_size(packet)) {
-        return EINVAL;
-    }
-
-    hlen = parse_gre_header(packet, tnl);
-    if (hlen < 0) {
-        return -hlen;
-    }
-
-    dp_packet_reset_packet(packet, hlen);
-
-    return 0;
-}
-
-static void
-netdev_gre_push_header(struct dp_packet *packet,
-                       const struct ovs_action_push_tnl *data)
-{
-    struct gre_base_hdr *greh;
-    int ip_tot_size;
-
-    greh = push_ip_header(packet, data->header, data->header_len, &ip_tot_size);
-
-    if (greh->flags & htons(GRE_CSUM)) {
-        ovs_be16 *csum_opt = (ovs_be16 *) (greh + 1);
-        *csum_opt = csum(greh, ip_tot_size);
-    }
-}
-
-static int
-netdev_gre_build_header(const struct netdev *netdev,
-                        struct ovs_action_push_tnl *data,
-                        const struct flow *tnl_flow)
-{
-    struct netdev_vport *dev = netdev_vport_cast(netdev);
-    struct netdev_tunnel_config *tnl_cfg;
-    struct ip_header *ip;
-    struct ovs_16aligned_ip6_hdr *ip6;
-    struct gre_base_hdr *greh;
-    ovs_16aligned_be32 *options;
-    int hlen;
-    bool is_ipv6;
-
-    is_ipv6 = is_header_ipv6(data->header);
-
-    /* XXX: RCUfy tnl_cfg. */
-    ovs_mutex_lock(&dev->mutex);
-    tnl_cfg = &dev->tnl_cfg;
-
-    if (is_ipv6) {
-        ip6 = ipv6_hdr(data->header);
-        ip6->ip6_nxt = IPPROTO_GRE;
-        greh = (struct gre_base_hdr *) (ip6 + 1);
-    } else {
-        ip = ip_hdr(data->header);
-        ip->ip_proto = IPPROTO_GRE;
-        greh = (struct gre_base_hdr *) (ip + 1);
-    }
-
-    greh->protocol = htons(ETH_TYPE_TEB);
-    greh->flags = 0;
-
-    options = (ovs_16aligned_be32 *) (greh + 1);
-    if (tnl_flow->tunnel.flags & FLOW_TNL_F_CSUM) {
-        greh->flags |= htons(GRE_CSUM);
-        put_16aligned_be32(options, 0);
-        options++;
-    }
-
-    if (tnl_cfg->out_key_present) {
-        greh->flags |= htons(GRE_KEY);
-        put_16aligned_be32(options, be64_to_be32(tnl_flow->tunnel.tun_id));
-        options++;
-    }
-
-    ovs_mutex_unlock(&dev->mutex);
-
-    hlen = (uint8_t *) options - (uint8_t *) greh;
-
-    data->header_len = sizeof(struct eth_header) + hlen +
-                       (is_ipv6 ? IPV6_HEADER_LEN : IP_HEADER_LEN);
-    data->tnl_type = OVS_VPORT_TYPE_GRE;
-    return 0;
-}
-
-static int
-netdev_vxlan_pop_header(struct dp_packet *packet)
-{
-    struct pkt_metadata *md = &packet->md;
-    struct flow_tnl *tnl = &md->tunnel;
-    struct vxlanhdr *vxh;
-    unsigned int hlen;
-
-    pkt_metadata_init_tnl(md);
-    if (VXLAN_HLEN > dp_packet_l4_size(packet)) {
-        return EINVAL;
-    }
-
-    vxh = udp_extract_tnl_md(packet, tnl, &hlen);
-    if (!vxh) {
-        return EINVAL;
-    }
-
-    if (get_16aligned_be32(&vxh->vx_flags) != htonl(VXLAN_FLAGS) ||
-       (get_16aligned_be32(&vxh->vx_vni) & htonl(0xff))) {
-        VLOG_WARN_RL(&err_rl, "invalid vxlan flags=%#x vni=%#x\n",
-                     ntohl(get_16aligned_be32(&vxh->vx_flags)),
-                     ntohl(get_16aligned_be32(&vxh->vx_vni)));
-        return EINVAL;
-    }
-    tnl->tun_id = htonll(ntohl(get_16aligned_be32(&vxh->vx_vni)) >> 8);
-    tnl->flags |= FLOW_TNL_F_KEY;
-
-    dp_packet_reset_packet(packet, hlen + VXLAN_HLEN);
-
-    return 0;
-}
-
-static int
-netdev_vxlan_build_header(const struct netdev *netdev,
-                          struct ovs_action_push_tnl *data,
-                          const struct flow *tnl_flow)
-{
-    struct netdev_vport *dev = netdev_vport_cast(netdev);
-    struct netdev_tunnel_config *tnl_cfg;
-    struct vxlanhdr *vxh;
-    unsigned int hlen;
-
-    /* XXX: RCUfy tnl_cfg. */
-    ovs_mutex_lock(&dev->mutex);
-    tnl_cfg = &dev->tnl_cfg;
-
-    vxh = udp_build_header(tnl_cfg, tnl_flow, data, &hlen);
-
-    put_16aligned_be32(&vxh->vx_flags, htonl(VXLAN_FLAGS));
-    put_16aligned_be32(&vxh->vx_vni, htonl(ntohll(tnl_flow->tunnel.tun_id) << 8));
-
-    ovs_mutex_unlock(&dev->mutex);
-    data->header_len = hlen + VXLAN_HLEN;
-    data->tnl_type = OVS_VPORT_TYPE_VXLAN;
-    return 0;
-}
-
-static int
-netdev_geneve_pop_header(struct dp_packet *packet)
-{
-    struct pkt_metadata *md = &packet->md;
-    struct flow_tnl *tnl = &md->tunnel;
-    struct genevehdr *gnh;
-    unsigned int hlen, opts_len, ulen;
-
-    pkt_metadata_init_tnl(md);
-    if (GENEVE_BASE_HLEN > dp_packet_l4_size(packet)) {
-        VLOG_WARN_RL(&err_rl, "geneve packet too small: min header=%u packet size=%"PRIuSIZE"\n",
-                     (unsigned int)GENEVE_BASE_HLEN, dp_packet_l4_size(packet));
-        return EINVAL;
-    }
-
-    gnh = udp_extract_tnl_md(packet, tnl, &ulen);
-    if (!gnh) {
-        return EINVAL;
-    }
-
-    opts_len = gnh->opt_len * 4;
-    hlen = ulen + GENEVE_BASE_HLEN + opts_len;
-    if (hlen > dp_packet_size(packet)) {
-        VLOG_WARN_RL(&err_rl, "geneve packet too small: header len=%u packet size=%u\n",
-                     hlen, dp_packet_size(packet));
-        return EINVAL;
-    }
-
-    if (gnh->ver != 0) {
-        VLOG_WARN_RL(&err_rl, "unknown geneve version: %"PRIu8"\n", gnh->ver);
-        return EINVAL;
-    }
-
-    if (gnh->proto_type != htons(ETH_TYPE_TEB)) {
-        VLOG_WARN_RL(&err_rl, "unknown geneve encapsulated protocol: %#x\n",
-                     ntohs(gnh->proto_type));
-        return EINVAL;
-    }
-
-    tnl->flags |= gnh->oam ? FLOW_TNL_F_OAM : 0;
-    tnl->tun_id = htonll(ntohl(get_16aligned_be32(&gnh->vni)) >> 8);
-    tnl->flags |= FLOW_TNL_F_KEY;
-
-    memcpy(tnl->metadata.opts.gnv, gnh->options, opts_len);
-    tnl->metadata.present.len = opts_len;
-    tnl->flags |= FLOW_TNL_F_UDPIF;
-
-    dp_packet_reset_packet(packet, hlen);
-
-    return 0;
-}
-
-static int
-netdev_geneve_build_header(const struct netdev *netdev,
-                           struct ovs_action_push_tnl *data,
-                           const struct flow *tnl_flow)
-{
-    struct netdev_vport *dev = netdev_vport_cast(netdev);
-    struct netdev_tunnel_config *tnl_cfg;
-    struct genevehdr *gnh;
-    int opt_len;
-    bool crit_opt;
-    unsigned int hlen;
-
-    /* XXX: RCUfy tnl_cfg. */
-    ovs_mutex_lock(&dev->mutex);
-    tnl_cfg = &dev->tnl_cfg;
-
-    gnh = udp_build_header(tnl_cfg, tnl_flow, data, &hlen);
-
-    put_16aligned_be32(&gnh->vni, htonl(ntohll(tnl_flow->tunnel.tun_id) << 8));
-
-    ovs_mutex_unlock(&dev->mutex);
-
-    opt_len = tun_metadata_to_geneve_header(&tnl_flow->tunnel,
-                                            gnh->options, &crit_opt);
-
-    gnh->opt_len = opt_len / 4;
-    gnh->oam = !!(tnl_flow->tunnel.flags & FLOW_TNL_F_OAM);
-    gnh->critical = crit_opt ? 1 : 0;
-    gnh->proto_type = htons(ETH_TYPE_TEB);
-
-    data->header_len = hlen + GENEVE_BASE_HLEN + opt_len;
-    data->tnl_type = OVS_VPORT_TYPE_GENEVE;
-    return 0;
-}
-
-static void
-netdev_vport_range(struct unixctl_conn *conn, int argc,
-                   const char *argv[], void *aux OVS_UNUSED)
-{
-    int val1, val2;
-
-    if (argc < 3) {
-        struct ds ds = DS_EMPTY_INITIALIZER;
-
-        ds_put_format(&ds, "Tunnel UDP source port range: %"PRIu16"-%"PRIu16"\n",
-                            tnl_udp_port_min, tnl_udp_port_max);
-
-        unixctl_command_reply(conn, ds_cstr(&ds));
-        ds_destroy(&ds);
-        return;
-    }
-
-    if (argc != 3) {
-        return;
-    }
-
-    val1 = atoi(argv[1]);
-    if (val1 <= 0 || val1 > UINT16_MAX) {
-        unixctl_command_reply(conn, "Invalid min.");
-        return;
-    }
-    val2 = atoi(argv[2]);
-    if (val2 <= 0 || val2 > UINT16_MAX) {
-        unixctl_command_reply(conn, "Invalid max.");
-        return;
-    }
-
-    if (val1 > val2) {
-        tnl_udp_port_min = val2;
-        tnl_udp_port_max = val1;
-    } else {
-        tnl_udp_port_min = val1;
-        tnl_udp_port_max = val2;
-    }
-    seq_change(tnl_conf_seq);
-
-    unixctl_command_reply(conn, "OK");
-}
-
 
 #define VPORT_FUNCTIONS(GET_CONFIG, SET_CONFIG,             \
                         GET_TUNNEL_CONFIG, GET_STATUS,      \
@@ -1495,7 +826,7 @@ netdev_vport_range(struct unixctl_conn *conn, int argc,
     PUSH_HEADER,                                            \
     POP_HEADER,                                             \
     NULL,                       /* get_numa_id */           \
-    NULL,                       /* set_multiq */            \
+    NULL,                       /* set_tx_multiq */         \
                                                             \
     NULL,                       /* send */                  \
     NULL,                       /* send_wait */             \
@@ -1527,15 +858,15 @@ netdev_vport_range(struct unixctl_conn *conn, int argc,
     NULL,                       /* queue_dump_done */       \
     NULL,                       /* dump_queue_stats */      \
                                                             \
-    NULL,                       /* get_in4 */               \
     NULL,                       /* set_in4 */               \
-    NULL,                       /* get_in6 */               \
+    NULL,                       /* get_addr_list */         \
     NULL,                       /* add_router */            \
     NULL,                       /* get_next_hop */          \
     GET_STATUS,                                             \
     NULL,                       /* arp_lookup */            \
                                                             \
     netdev_vport_update_flags,                              \
+    NULL,                       /* reconfigure */           \
                                                             \
     NULL,                   /* rx_alloc */                  \
     NULL,                   /* rx_construct */              \
@@ -1548,11 +879,12 @@ netdev_vport_range(struct unixctl_conn *conn, int argc,
 
 #define TUNNEL_CLASS(NAME, DPIF_PORT, BUILD_HEADER, PUSH_HEADER, POP_HEADER)   \
     { DPIF_PORT,                                                               \
-        { NAME, VPORT_FUNCTIONS(get_tunnel_config,                             \
-                                set_tunnel_config,                             \
-                                get_netdev_tunnel_config,                      \
-                                tunnel_get_status,                             \
-                                BUILD_HEADER, PUSH_HEADER, POP_HEADER) }}
+        { NAME, false,                                                         \
+          VPORT_FUNCTIONS(get_tunnel_config,                                   \
+                          set_tunnel_config,                                   \
+                          get_netdev_tunnel_config,                            \
+                          tunnel_get_status,                                   \
+                          BUILD_HEADER, PUSH_HEADER, POP_HEADER) }}
 
 void
 netdev_vport_tunnel_register(void)
@@ -1561,14 +893,14 @@ netdev_vport_tunnel_register(void)
      * a port number to the end if one is necessary. */
     static const struct vport_class vport_classes[] = {
         TUNNEL_CLASS("geneve", "genev_sys", netdev_geneve_build_header,
-                                            push_udp_header,
+                                            netdev_tnl_push_udp_header,
                                             netdev_geneve_pop_header),
         TUNNEL_CLASS("gre", "gre_sys", netdev_gre_build_header,
                                        netdev_gre_push_header,
                                        netdev_gre_pop_header),
         TUNNEL_CLASS("ipsec_gre", "gre_sys", NULL, NULL, NULL),
         TUNNEL_CLASS("vxlan", "vxlan_sys", netdev_vxlan_build_header,
-                                           push_udp_header,
+                                           netdev_tnl_push_udp_header,
                                            netdev_vxlan_pop_header),
         TUNNEL_CLASS("lisp", "lisp_sys", NULL, NULL, NULL),
         TUNNEL_CLASS("stt", "stt_sys", NULL, NULL, NULL),
@@ -1583,7 +915,7 @@ netdev_vport_tunnel_register(void)
         }
 
         unixctl_command_register("tnl/egress_port_range", "min max", 0, 2,
-                                 netdev_vport_range, NULL);
+                                 netdev_tnl_egress_port_range, NULL);
 
         ovsthread_once_done(&once);
     }
@@ -1594,9 +926,10 @@ netdev_vport_patch_register(void)
 {
     static const struct vport_class patch_class =
         { NULL,
-            { "patch", VPORT_FUNCTIONS(get_patch_config,
-                                       set_patch_config,
-                                       NULL,
-                                       NULL, NULL, NULL, NULL) }};
+            { "patch", false,
+              VPORT_FUNCTIONS(get_patch_config,
+                              set_patch_config,
+                              NULL,
+                              NULL, NULL, NULL, NULL) }};
     netdev_register_provider(&patch_class.netdev_class);
 }
