@@ -27,10 +27,10 @@
 
 #include "byte-order.h"
 #include "compiler.h"
-#include "dynamic-string.h"
+#include "openvswitch/dynamic-string.h"
 #include "netlink.h"
 #include "netlink-socket.h"
-#include "ofpbuf.h"
+#include "openvswitch/ofpbuf.h"
 #include "openvswitch/vlog.h"
 #include "poll-loop.h"
 #include "timeval.h"
@@ -118,7 +118,7 @@ struct nl_ct_dump_state {
     bool filter_zone;
     uint16_t zone;
 };
-
+
 /* Conntrack netlink dumping. */
 
 /* Initialize a conntrack netlink dump. */
@@ -200,7 +200,7 @@ nl_ct_dump_done(struct nl_ct_dump_state *state)
     free(state);
     return error;
 }
-
+
 /* Format conntrack event 'entry' of 'type' to 'ds'. */
 void
 nl_ct_format_event_entry(const struct ct_dpif_entry *entry,
@@ -235,6 +235,26 @@ nl_ct_flush(void)
     return err;
 }
 
+#ifdef _WIN32
+int
+nl_ct_flush_zone(uint16_t flush_zone)
+{
+    /* Windows can flush a specific zone */
+    struct ofpbuf buf;
+    int err;
+
+    ofpbuf_init(&buf, NL_DUMP_BUFSIZE);
+
+    nl_msg_put_nfgenmsg(&buf, 0, AF_UNSPEC, NFNL_SUBSYS_CTNETLINK,
+                        IPCTNL_MSG_CT_DELETE, NLM_F_REQUEST);
+    nl_msg_put_be16(&buf, CTA_ZONE, flush_zone);
+
+    err = nl_transact(NETLINK_NETFILTER, &buf, NULL);
+    ofpbuf_uninit(&buf);
+
+    return err;
+}
+#else
 int
 nl_ct_flush_zone(uint16_t flush_zone)
 {
@@ -299,7 +319,8 @@ nl_ct_flush_zone(uint16_t flush_zone)
      * have a master connection anymore */
     return 0;
 }
-
+#endif
+
 /* Conntrack netlink parsing. */
 
 static bool
@@ -476,7 +497,7 @@ nl_ct_parse_tuple(struct nlattr *nla, struct ct_dpif_tuple *tuple,
             struct ds ds;
 
             ds_init(&ds);
-            ct_dpif_format_tuple(&ds, tuple, true);
+            ct_dpif_format_tuple(&ds, tuple);
 
             VLOG_ERR_RL(&rl, "Failed to parse tuple: %s", ds_cstr(&ds));
             ds_destroy(&ds);
@@ -496,6 +517,10 @@ nl_ct_parse_tuple(struct nlattr *nla, struct ct_dpif_tuple *tuple,
 static uint8_t
 nl_ct_tcp_state_to_dpif(uint8_t state)
 {
+#ifdef _WIN32
+    /* Windows currently sends up CT_DPIF_TCP state */
+    return state;
+#else
     switch (state) {
     case TCP_CONNTRACK_NONE:
         return CT_DPIF_TCPS_CLOSED;
@@ -520,17 +545,23 @@ nl_ct_tcp_state_to_dpif(uint8_t state)
     default:
         return CT_DPIF_TCPS_CLOSED;
     }
+#endif
 }
 
 static uint8_t
 ip_ct_tcp_flags_to_dpif(uint8_t flags)
 {
+#ifdef _WIN32
+    /* Windows currently sends up CT_DPIF_TCP flags */
+    return flags;
+#else
     uint8_t ret = 0;
 #define CT_DPIF_TCP_FLAG(FLAG) \
         ret |= (flags & IP_CT_TCP_FLAG_##FLAG) ? CT_DPIF_TCPF_##FLAG : 0;
     CT_DPIF_TCP_FLAGS
 #undef CT_DPIF_STATUS_FLAG
     return ret;
+#endif
 }
 
 static bool
@@ -788,7 +819,7 @@ nl_ct_parse_entry(struct ofpbuf *buf, struct ct_dpif_entry *entry,
 
     return true;
 }
-
+
 /* NetFilter utility functions. */
 
 /* Puts a nlmsghdr and nfgenmsg at the beginning of 'msg', which must be
@@ -826,4 +857,8 @@ nl_msg_put_nfgenmsg(struct ofpbuf *msg, size_t expected_payload,
     nfm->nfgen_family = family;
     nfm->version = NFNETLINK_V0;
     nfm->res_id = 0;
+#ifdef _WIN32
+    /* nfgenmsg contains ovsHdr padding in windows */
+    nfm->ovsHdr.dp_ifindex = 0;
+#endif
 }
