@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2019 Nicira, Inc.
+ * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,16 +21,17 @@
 #include <stdlib.h>
 #include <string.h>
 #include "coverage.h"
-#include "ofp-msgs.h"
-#include "ofp-util.h"
-#include "ofpbuf.h"
 #include "openflow/openflow.h"
-#include "poll-loop.h"
-#include "sat-math.h"
-#include "timeval.h"
-#include "util.h"
+#include "openvswitch/ofp-msgs.h"
+#include "openvswitch/ofp-util.h"
+#include "openvswitch/ofpbuf.h"
 #include "openvswitch/vconn.h"
 #include "openvswitch/vlog.h"
+#include "poll-loop.h"
+#include "sat-math.h"
+#include "stream.h"
+#include "timeval.h"
+#include "util.h"
 
 VLOG_DEFINE_THIS_MODULE(rconn);
 
@@ -139,8 +140,7 @@ struct rconn {
     struct vconn *monitors[MAXIMUM_MONITORS];
     size_t n_monitors;
 
-    uint32_t allowed_versions;  /* Acceptable OpenFlow versions. */
-    int version;                /* Current or most recent version. */
+    uint32_t allowed_versions;
 };
 
 uint32_t rconn_get_allowed_versions(const struct rconn *rconn)
@@ -173,6 +173,8 @@ static bool is_connected_state(enum state);
 static bool is_admitted_msg(const struct ofpbuf *);
 static bool rconn_logging_connection_attempts__(const struct rconn *rc)
     OVS_REQUIRES(rc->mutex);
+static int rconn_get_version__(const struct rconn *rconn)
+    OVS_REQUIRES(rconn->mutex);
 
 /* The following prototypes duplicate those in rconn.h, but there we weren't
  * able to add the OVS_EXCLUDED annotations because the definition of struct
@@ -250,7 +252,7 @@ rconn_create(int probe_interval, int max_backoff, uint8_t dscp,
     rc->target = xstrdup("void");
     rc->reliable = false;
 
-    list_init(&rc->txq);
+    ovs_list_init(&rc->txq);
 
     rc->backoff = 0;
     rc->max_backoff = max_backoff ? max_backoff : 8;
@@ -273,9 +275,7 @@ rconn_create(int probe_interval, int max_backoff, uint8_t dscp,
     rconn_set_dscp(rc, dscp);
 
     rc->n_monitors = 0;
-
     rc->allowed_versions = allowed_versions;
-    rc->version = -1;
 
     return rc;
 }
@@ -340,6 +340,9 @@ rconn_connect(struct rconn *rc, const char *target, const char *name)
     rconn_disconnect__(rc);
     rconn_set_target__(rc, target, name);
     rc->reliable = true;
+    if (!stream_or_pstream_needs_probes(target)) {
+        rc->probe_interval = 0;
+    }
     reconnect(rc);
     ovs_mutex_unlock(&rc->mutex);
 }
@@ -364,7 +367,8 @@ rconn_connect_unreliably(struct rconn *rc,
     rconn_set_target__(rc, vconn_get_name(vconn), name);
     rc->reliable = false;
     rc->vconn = vconn;
-    state_transition(rc, S_CONNECTING);
+    rc->last_connected = time_now();
+    state_transition(rc, S_ACTIVE);
     ovs_mutex_unlock(&rc->mutex);
 }
 
@@ -489,7 +493,7 @@ static unsigned int
 timeout_CONNECTING(const struct rconn *rc)
     OVS_REQUIRES(rc->mutex)
 {
-    return MAX(2, rc->backoff);
+    return MAX(1, rc->backoff);
 }
 
 static void
@@ -498,10 +502,9 @@ run_CONNECTING(struct rconn *rc)
 {
     int retval = vconn_connect(rc->vconn);
     if (!retval) {
-        VLOG(rc->reliable ? VLL_INFO : VLL_DBG, "%s: connected", rc->name);
+        VLOG_INFO("%s: connected", rc->name);
         rc->n_successful_connections++;
         state_transition(rc, S_ACTIVE);
-        rc->version = vconn_get_version(rc->vconn);
         rc->last_connected = rc->state_entered;
     } else if (retval != EAGAIN) {
         if (rconn_logging_connection_attempts__(rc)) {
@@ -522,17 +525,17 @@ static void
 do_tx_work(struct rconn *rc)
     OVS_REQUIRES(rc->mutex)
 {
-    if (list_is_empty(&rc->txq)) {
+    if (ovs_list_is_empty(&rc->txq)) {
         return;
     }
-    while (!list_is_empty(&rc->txq)) {
+    while (!ovs_list_is_empty(&rc->txq)) {
         int error = try_send(rc);
         if (error) {
             break;
         }
         rc->last_activity = time_now();
     }
-    if (list_is_empty(&rc->txq)) {
+    if (ovs_list_is_empty(&rc->txq)) {
         poll_immediate_wake();
     }
 }
@@ -555,17 +558,19 @@ run_ACTIVE(struct rconn *rc)
 {
     if (timed_out(rc)) {
         unsigned int base = MAX(rc->last_activity, rc->state_entered);
+        int version;
+
         VLOG_DBG("%s: idle %u seconds, sending inactivity probe",
                  rc->name, (unsigned int) (time_now() - base));
 
+        version = rconn_get_version__(rc);
+        ovs_assert(version >= 0 && version <= 0xff);
+
         /* Ordering is important here: rconn_send() can transition to BACKOFF,
          * and we don't want to transition back to IDLE if so, because then we
          * can end up queuing a packet with vconn == NULL and then *boom*. */
         state_transition(rc, S_IDLE);
-
-        /* Send an echo request. */
-        rconn_send__(rc, make_echo_request(rc->version), NULL);
-
+        rconn_send__(rc, make_echo_request(version), NULL);
         return;
     }
 
@@ -671,7 +676,7 @@ rconn_run_wait(struct rconn *rc)
     ovs_mutex_lock(&rc->mutex);
     if (rc->vconn) {
         vconn_run_wait(rc->vconn);
-        if ((rc->state & (S_ACTIVE | S_IDLE)) && !list_is_empty(&rc->txq)) {
+        if ((rc->state & (S_ACTIVE | S_IDLE)) && !ovs_list_is_empty(&rc->txq)) {
             vconn_wait(rc->vconn, WAIT_SEND);
         }
     }
@@ -750,7 +755,7 @@ rconn_send__(struct rconn *rc, struct ofpbuf *b,
         /* Reuse 'frame' as a private pointer while 'b' is in txq. */
         b->header = counter;
 
-        list_push_back(&rc->txq, &b->list_node);
+        ovs_list_push_back(&rc->txq, &b->list_node);
 
         /* If the queue was empty before we added 'b', try to send some
          * packets.  (But if the queue had packets in it, it's because the
@@ -926,19 +931,23 @@ rconn_failure_duration(const struct rconn *rconn)
     return duration;
 }
 
-/* Returns the OpenFlow version most recently negotiated with a peer, or -1 if
- * no version has ever been negotiated.
- *
- * If 'rconn' is connected (that is, if 'rconn_is_connected(rconn)' would
- * return true), then the return value is guaranteed to be the OpenFlow version
- * in use for the connection.  The converse is not true: when the return value
- * is not -1, 'rconn' might be disconnected. */
+static int
+rconn_get_version__(const struct rconn *rconn)
+    OVS_REQUIRES(rconn->mutex)
+{
+    return rconn->vconn ? vconn_get_version(rconn->vconn) : -1;
+}
+
+/* Returns the OpenFlow version negotiated with the peer, or -1 if there is
+ * currently no connection or if version negotiation is not yet complete. */
 int
 rconn_get_version(const struct rconn *rconn)
     OVS_EXCLUDED(rconn->mutex)
 {
+    int version;
+
     ovs_mutex_lock(&rconn->mutex);
-    int version = rconn->version;
+    version = rconn_get_version__(rconn);
     ovs_mutex_unlock(&rconn->mutex);
 
     return version;
@@ -1000,7 +1009,7 @@ rconn_count_txqlen(const struct rconn *rc)
     unsigned int len;
 
     ovs_mutex_lock(&rc->mutex);
-    len = list_size(&rc->txq);
+    len = ovs_list_size(&rc->txq);
     ovs_mutex_unlock(&rc->mutex);
 
     return len;
@@ -1115,13 +1124,13 @@ try_send(struct rconn *rc)
     /* Eagerly remove 'msg' from the txq.  We can't remove it from the list
      * after sending, if sending is successful, because it is then owned by the
      * vconn, which might have freed it already. */
-    list_remove(&msg->list_node);
+    ovs_list_remove(&msg->list_node);
     msg->header = NULL;
 
     retval = vconn_send(rc->vconn, msg);
     if (retval) {
         msg->header = counter;
-        list_push_front(&rc->txq, &msg->list_node);
+        ovs_list_push_front(&rc->txq, &msg->list_node);
         if (retval != EAGAIN) {
             report_error(rc, retval);
             disconnect(rc, retval);
@@ -1216,11 +1225,11 @@ static void
 flush_queue(struct rconn *rc)
     OVS_REQUIRES(rc->mutex)
 {
-    if (list_is_empty(&rc->txq)) {
+    if (ovs_list_is_empty(&rc->txq)) {
         return;
     }
-    while (!list_is_empty(&rc->txq)) {
-        struct ofpbuf *b = ofpbuf_from_list(list_pop_front(&rc->txq));
+    while (!ovs_list_is_empty(&rc->txq)) {
+        struct ofpbuf *b = ofpbuf_from_list(ovs_list_pop_front(&rc->txq));
         struct rconn_packet_counter *counter = b->header;
         if (counter) {
             rconn_packet_counter_dec(counter, b->size);
@@ -1399,6 +1408,7 @@ is_admitted_msg(const struct ofpbuf *b)
     case OFPTYPE_ROLE_REPLY:
     case OFPTYPE_ROLE_STATUS:
     case OFPTYPE_REQUESTFORWARD:
+    case OFPTYPE_TABLE_STATUS:
     case OFPTYPE_SET_FLOW_FORMAT:
     case OFPTYPE_FLOW_MOD_TABLE_ID:
     case OFPTYPE_SET_PACKET_IN_FORMAT:
@@ -1415,6 +1425,12 @@ is_admitted_msg(const struct ofpbuf *b)
     case OFPTYPE_NXT_TLV_TABLE_MOD:
     case OFPTYPE_NXT_TLV_TABLE_REQUEST:
     case OFPTYPE_NXT_TLV_TABLE_REPLY:
+    case OFPTYPE_NXT_RESUME:
+    case OFPTYPE_IPFIX_BRIDGE_STATS_REQUEST:
+    case OFPTYPE_IPFIX_BRIDGE_STATS_REPLY:
+    case OFPTYPE_IPFIX_FLOW_STATS_REQUEST:
+    case OFPTYPE_IPFIX_FLOW_STATS_REPLY:
+    case OFPTYPE_CT_FLUSH_ZONE:
     default:
         return true;
     }
